use crate::statics::{
    *,
    format::*,
};
use crate::Component;

pub mod builder;

#[derive(Clone, Copy)]
pub enum Format {
    Single,
    Double,
    Half,
    Full,
}
impl Format {
    pub fn scale(&self) -> u8
    {
        match &self {
            Self::Single => 1,
            Self::Double => 2,
            Self::Half => 4,
            Self::Full => 8,
        }
    }

    pub fn size(&self) -> (u8, u8)
    {
        match &self {
            Self::Single => GRID_SINGLE,
            Self::Double => GRID_DOUBLE,
            Self::Half => GRID_HALF,
            Self::Full => GRID_FULL,
        }
    }
}

#[derive(Clone, Copy)]
pub enum Category {
    Equipment,
    Field,
    Ammunition,
    Structure,
    Unit,
    Upgrade,
    Weapon,
}

#[derive(Clone)]
pub struct Card {
    pub format:Format,
    category:Category,

    name:String,
    payload:u8,
    cost:u8,

    labels:Vec<usize>,

    components:[Vec<Component>; 2],
    grid:[Vec<u16>; 2],

    concealed:bool,
}
impl Card {
    pub fn new(format:Format) -> Self
    {
        let (cx, cy) = format.size();

        Self {
            format,
            category:Category::Ammunition,
            
            name:String::new(),
            payload:0,
            cost:0,

            labels:Vec::new(),

            components:[Vec::new(), Vec::new()],
            grid:[vec![0; (cx * cy) as usize], vec![0; (cx * cy) as usize]],

            concealed:false,
        }
    }

    pub fn builder(format:Format) -> builder::Builder
    {
        builder::Builder::new(Card::new(format))
    }

    pub fn add_component(&mut self, page:usize, component:Component) -> Result<(),()>
    {
        let coverage = component.coverage();
        let (cx, cy) = self.format.size();

        for cov in &coverage {
            let (x, y, _) = cov.clone();

            if x >= cx || y >= cy {
                return Err(());
            } else {
                let index = ((y * cx) + x) as usize;
                if self.grid[page][index] != 0 {
                    return Err(())
                }
            }
        }

        self.components[page].push(component);
        let id = self.components[page].len() as u16;
        for cov in &coverage {
            let (x, y, t) = cov.clone();

            let index = ((y * cx) + x) as usize;
            self.grid[page][index] = ((t as u16) << 8) | id;
        }
        Ok(())
    }

    pub fn generate_html(&self) -> (String, String)
    {
        let mut output = [String::new(), String::new()];

        output[0] = [
            String::from(format!("<article class=\"{}\">", self.css_class())),
            String::from("<table>"),
            self.generate_header(),
        ].concat();
        output[1] = [
            String::from(format!("<article class=\"{}\">", self.css_class())),
            String::from("<table>"),
            if !self.concealed { self.generate_header() } else { String::new() },
        ].concat();

        let (cx, cy) = self.format.size();
        for page in [0, 1] {
            let cy = if page == 0 { cy - 1 } else { cy };

            for y in 0..cy {
                output[page] += "<tr>";

                for x in 0..cx {
                    let cell = self.grid[page][((y * cx) + x) as usize];
                    if (page == 0 || !self.concealed) && cell != 0 {
                        let row = ((cell >> 8) & 0xFF) as u8;
                        if row != 0xFF {
                            output[page] += &self.components[page][(cell & 0xFF) as usize - 1].draw(row);
                        }
                    } else {
                        output[page] += "<td></td>";
                    }
                }

                output[page] += "</tr>";
            }
        }

        // Add labels to bottom of front page.
        output[0] += &format!("<tr><th class=\"right\" colspan=\"{}\">", cx - 1);
        for lbi in 0..self.labels.len() {
            output[0] += &LABELS[self.labels[lbi]].name;
            if lbi < self.labels.len() - 1 {
                output[0] += ", ";
            }
        }
        output[0] += "</th><td></td></tr>";

        // Terminate card tags.
        output[0] += "</table></article>";
        output[1] += "</table></article>";

        let [front, back] = output;
        (front, back)
    }

    fn generate_header(&self) -> String
    {
        let mut output = String::from("<tr>");
        let (cx, _) = self.format.size();

        output += &match self.category {
            Category::Field | Category::Upgrade => [
                format!("<th class=\"name\" colspan=\"{}\">{}</th>", cx - 1, self.name),
                format!("<th class=\"cost\">{}</th>", self.cost),
            ].concat(),
            _ => [
                format!("<th class=\"name\" colspan=\"{}\">{}</th>", cx - 2, self.name),
                format!("<th class=\"payload\">{}</th>", self.payload),
                format!("<th class=\"cost\">{}</th>", self.cost),
            ].concat(),
        };
        output += "</tr>";
        output
    }

    fn css_class(&self) -> String
    {
        match self.format {
            Format::Single => "single",
            Format::Double => "double",
            Format::Half => "half",
            Format::Full => "full",
        }.to_string() + " " + match &self.category {
            Category::Equipment => "equipment",
            Category::Field => "field",
            Category::Ammunition => "item",
            Category::Structure => "structure",
            Category::Unit => "unit",
            Category::Upgrade => "upgrade",
            Category::Weapon => "weapon",
        }
    }
}
