use std::f32::consts::TAU;

use crate::{statics::*, util};

mod field; use field::Field;
mod damage; use damage::Damage;
mod range; use range::Range;
mod ammo; use ammo::Ammo;
mod ammo_use; use ammo_use::AmmoUse;
mod hardpoint; use hardpoint::{Slots, Hardpoint};
mod health; use health::Health;
mod criticals; use criticals::Criticals;
mod rolltable; use rolltable::RollTable;

#[derive(Clone)]
pub enum ComponentData {
    Cell,
    Text(String),
    Field(Field),

    Damage(Damage),
    Range(Range),
    Ammo(Ammo),
    AmmoUse(AmmoUse),

    Armor(Health),
    Health(Health),
    Criticals(Criticals),

    Mount(Slots),
    Slots(Slots),

    RollTable(RollTable),

    Footer(String),
}

#[derive(Clone)]
pub struct Component {
    pub x:u8,
    pub y:u8,
    pub w:u8,
    pub h:u8,
    pub data:ComponentData,
}
impl Component {
    fn new() -> Self
    {
        Self {
            x:0,
            y:0,
            w:1,
            h:1,
            data:ComponentData::Cell,
        }
    }

    fn with(mut self, data:ComponentData) -> Self
    {
        self.data = data;
        self
    }

    pub fn cell() -> Self
    {
        Self::new()
    }

    pub fn text(text:&str) -> Self
    {
        Self::new().with(ComponentData::Text(text.to_string()))
    }

    pub fn field(field:&str, value:&str) -> Self
    {
        Self::new().with(ComponentData::Field(Field {
            field:field.to_string(),
            value:value.to_string(),
        }))
    }

    pub fn damage(damage:i8, penetration:i8, impact:i8, relative:bool) -> Self
    {
        Self::new().with(ComponentData::Damage(Damage {
            damage, penetration, impact, relative,
        }))
    }

    pub fn range(min:i8, short:i8, medium:i8, long:i8, max:i8, relative:bool) -> Self
    {
        Self::new().with(ComponentData::Range(Range {
            min, short, medium, long, max, relative,
        }))
    }

    pub fn ammo(index:Option<usize>, count:u8) -> Self
    {
        Self::new().with(ComponentData::Ammo(Ammo {
            index, count,
        }))
    }

    pub fn ammo_use(index:Option<usize>, hits:u8, salvo:u8) -> Self
    {
        Self::new().with(ComponentData::AmmoUse(AmmoUse {
            index, hits, salvo,
        }))
    }

    pub fn mount() -> Self
    {
        Self::new().with(ComponentData::Mount(Slots {
            hardpoints:Vec::new(),
        }))
    }

    pub fn slots() -> Self
    {
        Self::new().with(ComponentData::Slots(Slots {
            hardpoints:Vec::new(),
        }))
    }

    pub fn health(class:Option<u8>, entries:&[(&str, u8)]) -> Self
    {
        let mut ent = Vec::new();
        for (text, count) in entries {
            ent.push((text.to_string(), *count));
        }

        Self::new().with(ComponentData::Health(Health {
            class,
            entries:ent,
        }))
    }

    pub fn armor(class:Option<u8>, entries:&[(&str, u8)]) -> Self
    {
        let mut ent = Vec::new();
        for (text, count) in entries {
            ent.push((text.to_string(), *count));
        }

        Self::new().with(ComponentData::Armor(Health {
            class,
            entries:ent,
        }))
    }

    pub fn criticals(entries:&[(&str, u8)]) -> Self
    {
        let mut ent = Vec::new();
        for (text, count) in entries {
            ent.push((text.to_string(), *count));
        }

        Self::new().with(ComponentData::Criticals(Criticals {
            entries:ent,
        }))
    }

    pub fn rolltable(title:&str, entries:&[(&str, Option<u8>)]) -> Self
    {
        let mut ent = Vec::new();
        for (text, num) in entries {
            ent.push((text.to_string(), *num));
        }

        Self::new().with(ComponentData::RollTable(RollTable {
            title:title.to_string(),
            entries:ent,
        }))
    }


    pub fn hardpoint(mut self, class:usize, subclass:Option<usize>, size:u8, relative:Option<char>, internal:bool) -> Self
    {
        match &mut self.data {
            ComponentData::Mount(data) => {
                data.hardpoints.push(Hardpoint { class, subclass, size, relative, internal });
            }
            ComponentData::Slots(data) => {
                data.hardpoints.push(Hardpoint { class, subclass, size, relative, internal });
            }
            _ => { }
        }
        self
    }


    pub fn at(mut self, x:u8, y:u8) -> Self
    {
        self.x = x;
        self.y = y;
        self
    }

    pub fn scale(mut self, w:u8, h:u8) -> Self
    {
        self.w = w;
        self.h = h;
        self
    }

    pub fn coverage(&self) -> Vec<(u8, u8, u8)>
    {
        let mut result = Vec::new();
        match &self.data {
            ComponentData::Field(_) | ComponentData::Mount(_) | ComponentData::Slots(_) => {
                for y in 0..self.h {
                    for x in 0..self.w+1 {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
            }

            ComponentData::Damage(_) => {
                for y in 0..2 {
                    for x in 0..3 {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
                result.push((self.x, self.y + 1, 1));
            }

            ComponentData::Range(_) => {
                for y in 0..2 {
                    for x in 0..5 {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
                result.push((self.x, self.y + 1, 1));
            }

            ComponentData::Ammo(_) => {
                for y in 0..self.h {
                    for x in 0..3+self.w {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
            }

            ComponentData::AmmoUse(_) => {
                for y in 0..2 {
                    for x in 0..5 {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
                result.push((self.x, self.y + 1, 1));
            }

            ComponentData::Health(data) | ComponentData::Armor(data) => {
                let basis = if data.class.is_some() { 2 } else { 1 };
                for x in 0..basis+(data.entries.len() as u8 * self.w) {
                    for y in 0..self.h {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
            }

            ComponentData::Criticals(data) => {
                for y in 0..1+self.h {
                    for x in 0..1+(data.entries.len() as u8 * self.w) {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
                result.push((self.x, self.y + 1, 1));
            }

            ComponentData::RollTable(data) => {
                for y in 0..1+data.entries.len() as u8 {
                    for x in 0..5 {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                for y in 0..1+data.entries.len() as u8 {
                    result.push((self.x, self.y + y, y));
                }
            }

            _ => {
                for y in 0..self.h {
                    for x in 0..self.w {
                        result.push((self.x + x, self.y + y, 0xFF));
                    }
                }
                result.push((self.x, self.y, 0));
            }
        }
        result
    }

    pub fn draw(&self, row:u8) -> String
    {
        let mut output = String::new();
        match &self.data {
            ComponentData::Cell => {
                match row {
                    0 => {
                        output += &format!("<td class=\"color\" colspan=\"{}\" rowspan=\"{}\"></td>", self.w, self.h);
                    }
                    _ => { }
                }
            }

            ComponentData::Text(data) => {
                match row {
                    0 => {
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, data);
                    }
                    _ => { }
                }
            }

            ComponentData::Field(data) => {
                match row {
                    0 => {
                        output += &format!("<th rowspan=\"{}\">{}</th>", self.h, data.field);
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, data.value);
                    }
                    _ => { }
                }
            }

            ComponentData::Damage(data) => {
                match row {
                    0 => {
                        output += &format!("<th colspan=\"{}\">Dmg</th><th colspan=\"{}\">Pen</th><th colspan=\"{}\">Imp</th>", self.w, self.w, self.w);
                    }
                    1 => {
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.damage, false));
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.penetration, false));
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.impact, false));
                    }
                    _ => { }
                }
            }

            ComponentData::Range(data) => {
                match row {
                    0 => {
                        if data.relative {
                            output += &format!("<th colspan=\"{}\">Min</th>", self.w);
                            output += &format!("<th colspan=\"{}\">S</th>", self.w);
                            output += &format!("<th colspan=\"{}\">M</th>", self.w);
                            output += &format!("<th colspan=\"{}\">L</th>", self.w);
                            output += &format!("<th colspan=\"{}\">Max</th>", self.w);
                        }
                        else {
                            output += &format!("<th colspan=\"{}\">Min</th>", self.w);
                            output += &format!("<th colspan=\"{}\">S<sub>0</sub></th>", self.w);
                            output += &format!("<th colspan=\"{}\">M<sub>2</sub></th>", self.w);
                            output += &format!("<th colspan=\"{}\">L<sub>4</sub></th>", self.w);
                            output += &format!("<th colspan=\"{}\">Max<sub>6</sub></th>", self.w);
                        }
                    }
                    1 => {
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.min, data.relative));
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.short, data.relative));
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.medium, data.relative));
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.long, data.relative));
                        output += &format!("<td colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, util::print_value(data.max, data.relative));
                    }
                    _ => { }
                }
            }

            ComponentData::Ammo(data) => {
                match row {
                    0 => {
                        output += &format!("<th rowspan=\"{}\">Ammo</th>", self.h);
                        output += &format!("<td rowspan=\"{}\">{}</td>", self.h, util::print_ammo_class(data.index));
                        output += &format!("<td rowspan=\"{}\">{}</td>", self.h, util::print_value(data.count as i8, false));
                        output += &format!("<td class=\"bubble\" colspan=\"{}\" rowspan=\"{}\">{}</td>",
                            self.w,
                            self.h,
                            util::print_bubbles(data.count),
                        );
                    }
                    _ => { }
                }
            }

            ComponentData::AmmoUse(data) => {
                match row {
                    0 => {
                        output += "<th colspan=\"2\">Ammo</th>";
                        output += "<th>Hit</th>";
                        output += "<th></th>";
                        output += "<th>Salvo</th>";
                    }
                    1 => {
                        output += &format!("<td colspan=\"2\">{}</td>", util::print_ammo_class(data.index));
                        output += &format!("<td>{}</td>", util::print_value(data.hits as i8, false));
                        output += "<td></td>";
                        output += &format!("<td>{}</td>", util::print_value(data.salvo as i8, false));
                    }
                    _ => { }
                }
            }

            ComponentData::Mount(data) => {
                match row {
                    0 => {
                        let mut inner = String::new();
                        for hp in &data.hardpoints {
                            inner += "<div class=\"hardpoint\">";
                            inner += &format!("<div class=\"class\">{}</div>",
                                HARDPOINT_CODE[hp.class].name.to_string() + if hp.internal { "▾" } else { "" }
                            );
                            if let Some(subclass) = hp.subclass {
                                inner += &format!("<div class=\"subclass\">{}</div>", HARDPOINT_SUBCODE[subclass].name);
                            }
                            inner += &format!("<div class=\"class\">{}</div>", if let Some(rel) = hp.relative {
                                if hp.size > 0 {
                                    format!("{}+{}", rel, hp.size)
                                } else {
                                    format!("{}", rel)
                                }
                            } else { format!("{}", hp.size) });
                            inner += "</div>";
                        }

                        output += &format!("<th rowspan=\"{}\">Mount</th>", self.h);
                        output += &format!("<td class=\"left\" colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, inner);
                    }
                    _ => { }
                }
            }

            ComponentData::Slots(data) => {
                match row {
                    0 => {
                        let mut inner = String::new();
                        for hp in &data.hardpoints {
                            inner += "<div class=\"hardpoint\">";
                            inner += &format!("<div class=\"class\">{}</div>",
                                HARDPOINT_CODE[hp.class].name.to_string() + if hp.internal { "▾" } else { "" }
                            );
                            if let Some(subclass) = hp.subclass {
                                inner += &format!("<div class=\"subclass\">{}</div>", HARDPOINT_SUBCODE[subclass].name);
                            }
                            inner += &format!("<div class=\"class\">{}</div>", if let Some(rel) = hp.relative {
                                if hp.size > 0 {
                                    format!("{}+{}", rel, hp.size)
                                } else {
                                    format!("{}", rel)
                                }
                            } else { format!("{}", hp.size) });
                            inner += "</div>";
                        }

                        output += &format!("<th rowspan=\"{}\">Slots</th>", self.h);
                        output += &format!("<td class=\"left\" colspan=\"{}\" rowspan=\"{}\">{}</td>", self.w, self.h, inner);
                    }
                    _ => { }
                }
            }

            ComponentData::Armor(data) | ComponentData::Health(data) => {
                match row {
                    0 => {
                        output += &format!("<th rowspan=\"{}\">{}</th>", self.h, match &self.data {
                            ComponentData::Armor(_) => LANG_ARMOR,
                            ComponentData::Health(_) => LANG_HEALTH,
                            _ => "",
                        });

                        if let Some(class) = data.class {
                            output += &format!("<td rowspan=\"{}\">{}</td>", self.h, util::print_value(class as i8, false));
                        }

                        for (label, count) in &data.entries {
                            output += &format!("<td class=\"bubble\" colspan=\"{}\" rowspan=\"{}\">", self.w, self.h);
                            output += &format!("<div class=\"corner\">{}</div>", label);
                            output += &util::print_bubbles(*count);
                            output += "</td>";
                        }
                    }
                    _ => { }
                }
            }

            ComponentData::Criticals(data) => {
                match row {
                    0 => {
                        output += &format!("<th></th>");

                        for (text, _) in &data.entries {
                            output += &format!("<th colspan=\"{}\">{}</th>", self.w, text);
                        }
                    }
                    1 => {
                        output += &format!("<th rowspan=\"{}\">{}</th>", self.h, LANG_CRITICALS);
                        for (_, count) in &data.entries {
                            output += &format!("<td class=\"bubble\" colspan=\"{}\" rowspan=\"{}\">", self.w, self.h);
                            output += &util::print_bubbles(*count);
                            output += "</td>";
                        }
                    }
                    _ => { }
                }
            }

            ComponentData::RollTable(data) => {
                match row { 
                    0 => {
                        output += &format!("<th colspan=\"5\">{}</th>", data.title);
                    }
                    _ => {
                        let (text, num) = &data.entries[row as usize - 1];
                        output += &format!("<th>{}</th>", row + 1);
                        output += &format!("<td colspan=\"3\">{}</td>", text);
                        if let Some(num) = num {
                            output += &format!("<td>{}</td>", util::print_size(*num as i8));
                        } else {
                            output += "<td></td>";
                        }
                    }
                }
            }

            _ => { }
        }
        output
    }
}
